/* ******************************************************* */
/* This module is a port of John Shalf's HDF4 reader for   */ 
/* AVS5.                                                   */
/* Cristina                                                */
/* ******************************************************* */ 
/* #include "../../user.h"
*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <avs/err.h>
#include <avs/om.h>

#include "sds.h"

typedef struct SDSinfo {
  int sdsnumber;
  int32 numbertype,elementsize;
  int32 rank,dims[5];
  char *Name;
  char *Label;
  char *Units;
  char *Annotation;
  char *WidgetLabel;
}SDSinfo;

int
HDF4reader(OMobj_id readHDF_id, OMevent_mask event_mask, int seq_num)
{
   /***********************/
   /*  Declare variables  */
   /***********************/
   char  *filename = NULL;
   int  field_number;
   int  data_layout;
   int  debug;
   OMobj_id output_id;
   OMobj_id param_id, param_id_field, param_id_number;
   int output_ndim, *output_dims, output_nspace, output_npoints;
   float *output_points;
   int output_ncomp, output_comp_count, output_veclen;
   int    output_data_type, output_ndata;
   char  *output_data;
	int max_field;

   /***********************/
	/* User Varaiables     */
   /***********************/
	static int firstcall=1, nSelectable=0;
	static char *choice_str;
	static char currentfile[256];
	static int nsds, last_field_number=-2;
	static SDSinfo *sdsinfo;
	static void *field, *field_buffer1;
	char *numberTypeString;
	register int i;
	int status;
	int value;
	char typeNameBuffer[256];
			
   /***********************/
   /*  Get input values   */
   /***********************/
   /* Get filename's value */
   if (OMget_name_str_val(readHDF_id, OMstr_to_name("filename"), &filename, 0) != 1)
      filename = NULL;

   /* Get field_number's value */ 
   if (OMget_name_int_val(readHDF_id, OMstr_to_name("field_number"), &field_number) != 1) 
      field_number = 0;

   /* Get data_layout's value */
   if (OMget_name_int_val(readHDF_id, OMstr_to_name("data_layout"), &data_layout) != 1)
      data_layout = 1; /* F77 */

   /* Get debug's value */ 
   if (OMget_name_int_val(readHDF_id, OMstr_to_name("debug"), &debug) != 1) 
      debug = 0;


   /***********************/
   /* Function's Body     */
   /***********************/
   ERRverror("",ERR_NO_HEADER | ERR_INFO,"I'm in function: readHDF4 generated from method: readHDF4.readHDF4\n");

   /***********************/
	/* User code           */
   /***********************/

	if((filename==NULL) || (strlen(filename)==0))
		return 1;
	if(firstcall) {
		sdsinfo=NULL;
		*currentfile='\0';
		firstcall=0;
	};

	param_id = OMfind_subobj(readHDF_id, OMstr_to_name("filename"), OM_OBJ_RD);	
	if (OMchanged(param_id, seq_num) || sdsinfo == NULL){
		int n;
		if (strlen(currentfile))
			SDSclose(currentfile);
		strcpy(currentfile, filename);
		SDSopen(currentfile, "r");
		if (debug)
			printf("currentfile=%s\n", currentfile);

		if (sdsinfo) {

			for (i=0; i < nsds; i++){
				if (sdsinfo[i].Name)
					free (sdsinfo[i].Name);
				if(sdsinfo[i].Label)
					free (sdsinfo[i].Label);
				if(sdsinfo[i].Units)
					free (sdsinfo[i].Units);
				if(sdsinfo[i].Annotation)
					free(sdsinfo[i].Annotation);
				if(sdsinfo[i].WidgetLabel)
					free(sdsinfo[i].WidgetLabel);
			}
			free(sdsinfo);
			sdsinfo = NULL;
		}
		nsds = SDSgetNumDatasets(filename);
		if(debug)
			printf("Allocating nsds = %u\n", nsds);
		sdsinfo = (SDSinfo *) calloc(nsds, sizeof(SDSinfo));
		/* go from first to last in file */
		for(i=0,n=0; i < nsds; i++){
			int t;
			SDSseek(filename,i);
			if(SDSisCoord(filename))
				continue;
			SDSgetDims(filename, typeNameBuffer, &(sdsinfo[n].rank), &(sdsinfo[n].dims[0]));
			if(debug){
				printf("%3u,%3u: rank=%u\n",i,n,sdsinfo[n].rank);
				for (t=0; t<sdsinfo[n].rank; t++){
					printf("\tDims[%u]=%u\n",t,sdsinfo[n].dims[t]);
				}
			}
			sdsinfo[n].numbertype = SDSgetNT(filename);
			sdsinfo[n].sdsnumber = i;
			n++;	
		}
		nSelectable = n;
		if (field_number > nSelectable) {
			param_id_number = OMfind_subobj(readHDF_id, OMstr_to_name("field_number"), OM_OBJ_RD);
			value = nSelectable - 1;
			field_number = value;
        	OMset_int_val(param_id_number, value);
		}	
	}
	
	if(!sdsinfo) return 1;/* go home, nuthin to see here....*/	

	switch (sdsinfo[field_number].numbertype) {

	case DFNT_FLOAT32:
		if (debug)
			puts("Its float");
		numberTypeString="float";
		output_data_type = DTYPE_FLOAT;	
		break;
	case DFNT_FLOAT64:
		if (debug)
			puts("Its double");
		numberTypeString="double";
		output_data_type = DTYPE_DOUBLE;	
		break;
   case DFNT_INT8:
      if (debug)
         puts("Its byte");
      numberTypeString="byte";
		output_data_type = DTYPE_BYTE;	
      break;
   case DFNT_INT16:
      if (debug)
         puts("Its short");
      numberTypeString="short";
		output_data_type = DTYPE_SHORT;	
      break;
   case DFNT_INT32:
      if (debug)
         puts("Its int");
      numberTypeString="int";
		output_data_type = DTYPE_INT;	
      break;
   case DFNT_CHAR:
      if (debug)
         puts("Its char");
      numberTypeString="char";
		output_data_type = DTYPE_CHAR;	
      break;
	default: /* error */
		printf("I don't recognize this datatype DFNT_???? number=[%u]\n",
			sdsinfo[field_number].numbertype);
		numberTypeString="";
		return 1;
		break;
	}	

	sprintf(typeNameBuffer,"field %uD %u-space uniform scalar %s",
		sdsinfo[field_number].rank,sdsinfo[field_number].rank,numberTypeString);

	if (debug) {
		puts(typeNameBuffer);
		for(i=0;i<sdsinfo[field_number].rank && i<5;i++)
			printf("Dims[%u]=%u\n",i,sdsinfo[field_number].dims[i]);
	}	

   /***********************/
   /*  Set output values  */
   /***********************/
	/* Set max_field */
	param_id_field = OMfind_subobj(readHDF_id, OMstr_to_name("max_field"), OM_OBJ_RW);
  	OMset_int_val(param_id_field, nSelectable - 1);
	
   /* Set output uniform mesh */

   /*  Get mesh id */
   output_id = OMfind_subobj(readHDF_id, OMstr_to_name("output"), OM_OBJ_RW);

   /* Set mesh dimensionality, output_ndim can be 1,2 or 3 */
	output_ndim = sdsinfo[field_number].rank;
   FLDset_ndim (output_id, output_ndim);

   /* Set mesh dims array */
   output_dims = (int *)ARRalloc(NULL, DTYPE_INT, 
                           output_ndim, NULL);

   /*** fill in dims array with your values ***/
	for ( i = 0; i < output_ndim; i++) {
		if (data_layout == 1) /* F77 */	
			output_dims[i] = sdsinfo[field_number].dims[output_ndim - 1 - i];
		else if (data_layout == 0) 
			output_dims[i] = sdsinfo[field_number].dims[i];
		else {
			printf("I don't recognize this data_layout\n");	
			return 1;
		}
	}
   FLDset_dims (output_id, output_dims);
   if (output_dims)
      ARRfree(output_dims);

   /* Set mesh nspace, output_nspace can be 1,2 or 3 */
	output_nspace = output_ndim;
   FLDset_nspace (output_id, output_nspace);

   /* Set mesh extents */
   output_npoints = 2*output_nspace;
   output_points = (float *)ARRalloc(NULL, DTYPE_FLOAT, 
                               output_npoints, NULL);

   /*** fill in points array with values for 2 points: low left and high right corners of the mesh ***/

	for (i=0; i < output_nspace; i++){
		output_points[i] = 0;
		output_points[i+output_nspace] = sdsinfo[field_number].dims[i];
	}
   FLDset_points (output_id, output_points, output_npoints, 
                  OM_SET_ARRAY_COPY);
   /* set  output node data */

   /* Get field id */
   output_id = OMfind_subobj(readHDF_id, OMstr_to_name("output"), OM_OBJ_RW);

   /* Set number of node data components */
	output_ncomp = 1;
   FLDset_node_data_ncomp (output_id, output_ncomp);

   /* For each node data component set veclen, type and data arry itself */
   for (output_comp_count=0; output_comp_count < output_ncomp; output_comp_count++) {

      /* Set veclen, assign output_veclen before next call */
		output_veclen = 1;
      FLDset_node_data_veclen (output_id,output_comp_count,output_veclen);

      /* Set data array */
      /* data_type should be set to one of the following: 
         DTYPE_BYTE, DTYPE_CHAR, DTYPE_SHORT, 
         DTYPE_INT, DTYPE_FLOAT, DTYPE_DOUBLE) */


      /* allocate output_data array first */
      /* assume float array and output_ndata is set to number of nodes */
		for (i = 0, output_ndata = 1; i < sdsinfo[field_number].rank; i++)
			output_ndata *= (sdsinfo[field_number].dims[i]);

      output_data = (char *)ARRalloc(NULL, output_data_type, 
                                 output_ndata, NULL);
		
	   /* OK, now SDSseek to proper field */
   	if(debug)
      	printf("seek to SDS %u\n",sdsinfo[field_number].sdsnumber);
   	SDSseek(filename,sdsinfo[field_number].sdsnumber);
   	/* already know the dims, so just read it */
   	SDSreadData(filename,(sdsinfo[field_number].rank),
      	&(sdsinfo[field_number].dims[0]),(char *) output_data); 

      FLDset_node_data (output_id, output_comp_count, output_data, output_data_type,
                        output_veclen*output_ndata, OM_SET_ARRAY_COPY);
      /*  Other useful calls:

          FLDset_node_data_id()
          FLDset_node_null_data()
          FLDset_node_data_label()
      */
   }
   

   /*************************/
   /*  Free input variables */
   /*************************/
   if (filename)
      free(filename);
	if (output_points)
		free(output_points);
	if (output_data)
		free(output_data);

   return(1);
}
